import {
	BadRequestException,
	Injectable,
	UnauthorizedException
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { JWTToken, tfaTypes, User, UserRole } from '../entities';
import { LessThanOrEqual, Repository } from 'typeorm';
import * as argon2 from 'argon2';
import { PassportStrategy } from '@nestjs/passport';
import { Strategy as LocalStrategy } from 'passport-local';
import { ExtractJwt, Strategy as JWTStrategy } from 'passport-jwt';
import FileSystemService from './filesystem';
import * as jwt from 'jsonwebtoken';
import { createTransport } from 'nodemailer';
import * as notp from 'notp';
import { randomBytes } from 'crypto';

const jwtSecret = 'CUM';

const mailAccount = createTransport({
	host: 'smtp.web.de',
	port: 587,
	secure: false,
	auth: require('M:/projects/file_server/dist/auth.json')
});

interface jwtPayload {
	sub: number;
	jti: number;
	exp?: number;
	iat?: number;
}

@Injectable()
export class AuthService {
	constructor(
		@InjectRepository(User)
		private userRepo: Repository<User>,
		@InjectRepository(JWTToken)
		private tokenRepo: Repository<JWTToken>,
		private fsService: FileSystemService
	) {}

	generateTfaSecret(): string {
		const set =
			'0123456789ABCDEFGHIJKLMNOPQRSTUVWXTZabcdefghiklmnopqrstuvwxyz!@#$%^&*()<>?/[]{},.:;';
		return randomBytes(32)
			.map((b) =>
				set.charCodeAt(Math.floor((b / 255.0) * (set.length - 1)))
			)
			.toString();
	}

	async getUser(userId: number): Promise<User | null> {
		return this.userRepo.findOneBy({
			id: userId
		});
	}

	async findUser(username: string): Promise<User | null> {
		return this.userRepo.findOneBy({
			name: username
		});
	}

	async getToken(tokenId: number): Promise<JWTToken | null> {
		return this.tokenRepo.findOneBy({
			id: tokenId
		});
	}

	async validateUser(username: string, pass: string): Promise<User | null> {
		const user = await this.findUser(username);
		if (!user)
			throw new UnauthorizedException('Invalid username or password');
		if (!(await argon2.verify(user.password, pass)))
			throw new UnauthorizedException('Invalid username or password');
		if (user.role == UserRole.DISABLED)
			throw new UnauthorizedException('Account is disabled');
		return user;
	}

	async cleanupTokens(): Promise<void> {
		await this.tokenRepo.delete({
			exp: LessThanOrEqual(Math.floor(Date.now() / 1000))
		});
	}

	requiresTfa(user: User): boolean {
		return user.tfaType != tfaTypes.NONE;
	}

	async verifyTfa(
		user: User,
		token: string,
		type?: tfaTypes
	): Promise<boolean> {
		if (!type) type = user.tfaType;
		const delta = notp.totp.verify(token, user.tfaSecret, {
			window: 10
		});
		return (
			delta &&
			(type == tfaTypes.EMAIL ? delta.delta <= 0 : delta.delta == 0)
		);
	}

	async sendTfaMail(user: User) {
		await mailAccount.sendMail({
			from: 'matthiasveigel@web.de',
			to: user.name,
			subject: 'Fileserver - EMail 2fa code',
			text: `Your code is: ${notp.totp.gen(
				user.tfaSecret
			)}\nIt is valid for 5 Minutes`
		});
	}

	async setupTfa(user: User): Promise<string> {
		if (user.tfaType != tfaTypes.NONE)
			throw new BadRequestException(
				'2 Factor authentication is already setup'
			);
		const secret = this.generateTfaSecret();
		user.tfaSecret = secret;
		await this.userRepo.save(user);
		return secret;
	}

	async setTfaType(user: User, type: tfaTypes) {
		user.tfaType = type;
		await this.userRepo.save(user);
	}

	async login(user: User) {
		const token = new JWTToken();
		token.ownerId = user.id;
		const db_token = await this.tokenRepo.save(token);
		const payload: jwtPayload = {
			sub: user.id,
			jti: db_token.id
		};
		const jwtToken = jwt.sign(payload, jwtSecret, {
			mutatePayload: true,
			expiresIn: '1d'
		});
		db_token.exp = payload.exp;
		await this.tokenRepo.save(db_token);
		return jwtToken;
	}

	async signup(username: string, password: string) {
		if (await this.findUser(username))
			throw new BadRequestException('User already exists');
		const user = new User();
		user.name = username;
		user.password = await argon2.hash(password);
		const dbUser = await this.userRepo.save(user);
		const root = await this.fsService.generateRoot(dbUser);
		dbUser.rootId = root.id;
		await this.userRepo.save(dbUser);
	}

	async revoke(token: JWTToken) {
		await this.tokenRepo.delete({
			id: token.id
		});
	}

	async revokeAll(user: User) {
		await this.tokenRepo.delete({
			ownerId: user.id
		});
	}
}

@Injectable()
export class AuthLocalService extends PassportStrategy(LocalStrategy) {
	constructor(private authService: AuthService) {
		super();
	}

	async validate(username: string, pass: string) {
		const user = await this.authService.validateUser(username, pass);
		if (!user)
			throw new UnauthorizedException('Invalid username or password');
		return user;
	}
}

@Injectable()
export class AuthJwtService extends PassportStrategy(JWTStrategy) {
	constructor(private authService: AuthService) {
		super({
			jwtFromRequest: ExtractJwt.fromExtractors([
				ExtractJwt.fromAuthHeaderAsBearerToken(),
				ExtractJwt.fromBodyField('jwtToken')
			]),
			ignoreExpiration: false,
			passReqToCallback: true,
			secretOrKey: jwtSecret
		});
	}

	async validate(req: Request, payload: jwtPayload) {
		await this.authService.cleanupTokens();
		const token = await this.authService.getToken(payload.jti);
		if (!token)
			throw new UnauthorizedException(
				'Invalid token, please log in again'
			);
		const user = await this.authService.getUser(token.ownerId);
		if (!user || user.id != payload.sub)
			throw new UnauthorizedException(
				'Invalid token, please log in again'
			);
		// eslint-disable-next-line @typescript-eslint/ban-ts-comment
		// @ts-ignore
		req.token = token;
		return user;
	}
}
