import {
	BadRequestException,
	Injectable,
	NotImplementedException,
	StreamableFile,
	UnauthorizedException
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { INode, User } from '../entities';
import { Repository } from 'typeorm';
import { Multipart } from '@fastify/multipart';
import { pipeline } from 'stream/promises';
import { createReadStream, createWriteStream, statSync, unlink } from 'fs';
import { Writable } from 'stream';

@Injectable()
export default class FileSystemService {
	constructor(
		@InjectRepository(INode)
		private inodeRepo: Repository<INode>
	) {}

	async generateRoot(user: User): Promise<INode> {
		const node = new INode();
		node.isFile = false;
		node.name = '';
		node.owner = Promise.resolve(user);
		return await this.inodeRepo.save(node);
	}

	async getNode(nodeId: number): Promise<INode> {
		return await this.inodeRepo.findOneBy({
			id: nodeId
		});
	}

	async getNodeAndValidate(node_id: number, user: User): Promise<INode> {
		const node = await this.getNode(node_id);
		if (node == null) throw new BadRequestException();
		if (node.ownerId != user.id) throw new UnauthorizedException();
		return node;
	}

	async generatePath(node: INode): Promise<string> {
		if (node.parentId == null) return '/';
		return (
			(await this.generatePath(await node.parent)).slice(0, -1) +
			'/' +
			node.name +
			(node.isFile ? '' : '/')
		);
	}

	async delete(node: INode): Promise<void> {
		if (node.parentId == null)
			throw new BadRequestException("Can't delete root");
		if (!node.isFile)
			await Promise.all(
				(await node.children).map((child) => this.delete(child))
			);
		else
			unlink(`files/${node.id}`, (err) => {
				if (err) console.error(`Error while deleting ${node.id}`, err);
			});
		await this.inodeRepo.remove(node);
	}

	async create(
		parent: INode,
		full_name: string,
		owner: User,
		file: boolean
	): Promise<INode> {
		const name = full_name.trim();
		if (name == '') throw new BadRequestException("Name can't be empty");
		if (name == '.' || name == '..')
			throw new BadRequestException('Invalid name');
		if (parent.isFile)
			throw new BadRequestException("Can't create file/folder in file");
		if (
			await this.inodeRepo.findOneBy({
				parentId: parent.id,
				name: name
			})
		)
			throw new BadRequestException('File/Folder already exists');
		const node = new INode();
		node.isFile = file;
		node.name = name;
		node.owner = Promise.resolve(owner);
		node.parent = Promise.resolve(parent);
		return await this.inodeRepo.save(node);
	}

	async uploadFile(file: Multipart, nodeId: number, user: User) {
		try {
			const node = await this.getNodeAndValidate(nodeId, user);

			await pipeline(file.file, createWriteStream(`files/${node.id}`));

			const stats = statSync(`files/${node.id}`);
			node.size = stats.size;
			await this.inodeRepo.save(node);
		} catch (e) {
			await pipeline(
				file.file,
				new Writable({
					write(
						chunk: any,
						encoding: BufferEncoding,
						callback: (error?: Error | null) => void
					) {
						setImmediate(callback);
					}
				})
			);
		}
	}

	async downloadFile(id: number, user: User): Promise<StreamableFile> {
		const node = await this.getNodeAndValidate(id, user);
		if (!node.isFile) throw new NotImplementedException();
		const stats = statSync(`files/${node.id}`);
		return new StreamableFile(createReadStream(`files/${node.id}`), {
			disposition: `attachment; filename="${node.name}"`,
			length: stats.size
		});
	}
}
